import CoreImage
import TensorFlowLite
import UIKit
import Accelerate

struct Result {
    let inferenceTime: Double
    let inferences: [Inference]
}

struct Inference {
    let confidence: Float
    let label: String
}

typealias FileInfo = (name: String, extension: String)
enum MobileNet {
    static let livenessModel: FileInfo = (name: "liveness", extension: "tflite")
    static let cardModel: FileInfo = (name: "card", extension: "tflite")
    static let faceModel: FileInfo = (name: "face", extension: "tflite")
}

class SBKModelDataHandler {
    let threadCount: Int
    let resultCount = 3
    let threadCountLimit = 10
    var subtract: Float = 127.5
    
    // MARK: - Model Parameters
    
    let batchSize = 1
    let inputChannels = 3
    var inputWidth = 256
    var inputHeight = 256
    private var labels: [String] = []
    private var interpreter: Interpreter
    private let alphaComponent = (baseOffset: 4, moduloRemainder: 3)
    init?(modelFileInfo: FileInfo, threadCount: Int = 1) {
        let modelFilename = modelFileInfo.name
        let bundle = Bundle(for: type(of: self))
        guard let modelPath = bundle.path(
            forResource: modelFilename,
            ofType: modelFileInfo.extension
        ) else {
            print("Failed to load the model file with name: \(modelFilename).")
            return nil
        }
        self.threadCount = threadCount
        var options = Interpreter.Options()
        options.threadCount = threadCount
        do {
            interpreter = try Interpreter(modelPath: modelPath, options: options)
            try interpreter.allocateTensors()
        } catch let error {
            print("Failed to create the interpreter with error: \(error.localizedDescription)")
            return nil
        }
    }
    
    func runModel(onFrame pixelBuffer: CVPixelBuffer) -> [Float]? {
      
      let sourcePixelFormat = CVPixelBufferGetPixelFormatType(pixelBuffer)
      assert(sourcePixelFormat == kCVPixelFormatType_32ARGB ||
               sourcePixelFormat == kCVPixelFormatType_32BGRA ||
                 sourcePixelFormat == kCVPixelFormatType_32RGBA)


      let imageChannels = 4
      assert(imageChannels >= inputChannels)

      // Crops the image to the biggest square in the center and scales it down to model dimensions.
      let scaledSize = CGSize(width: inputWidth, height: inputHeight)
      guard let thumbnailPixelBuffer = pixelBuffer.resized(to: scaledSize) else {
        return nil
      }
      let interval: TimeInterval
      let outputTensor: Tensor
      do {
        let inputTensor = try interpreter.input(at: 0)

        // Remove the alpha component from the image buffer to get the RGB data.
        guard let rgbData = rgbDataFromBuffer(
          thumbnailPixelBuffer,
          byteCount: batchSize * inputWidth * inputHeight * inputChannels,
          isModelQuantized: inputTensor.dataType == .uInt8
        ) else {
          print("Failed to convert the image buffer to RGB data.")
          return nil
        }

          let imageCap = UIImage(data: rgbData)
         // self.fromImage(image: imageCap!, datas: rgbData, imagesss: imageCap!)
          
        // Copy the RGB data to the input `Tensor`.
        try interpreter.copy(rgbData, toInputAt: 0)

        // Run inference by invoking the `Interpreter`.
        let startDate = Date()
        try interpreter.invoke()
        interval = Date().timeIntervalSince(startDate) * 1000

        // Get the output `Tensor` to process the inference results.
        outputTensor = try interpreter.output(at: 0)
      } catch let error {
        print("Failed to invoke the interpreter with error: \(error.localizedDescription)")
        return nil
      }

      let results: [Float]
      switch outputTensor.dataType {
      case .uInt8:
        guard let quantization = outputTensor.quantizationParameters else {
          print("No results returned because the quantization values for the output tensor are nil.")
          return nil
        }
        let quantizedResults = [UInt8](outputTensor.data)
        results = quantizedResults.map {
          quantization.scale * Float(Int($0) - quantization.zeroPoint)
        }
      case .float32:
        results = [Float32](unsafeData: outputTensor.data) ?? []
      default:
        print("Output tensor data type \(outputTensor.dataType) is unsupported for this example app.")
        return nil
      }

      return results
    }
    
    
    private func rgbDataFromBuffer(
        _ buffer: CVPixelBuffer,
        byteCount: Int,
        isModelQuantized: Bool
    ) -> Data? {
        CVPixelBufferLockBaseAddress(buffer, .readOnly)
        defer {
            CVPixelBufferUnlockBaseAddress(buffer, .readOnly)
        }
        guard let sourceData = CVPixelBufferGetBaseAddress(buffer) else {
            return nil
        }
        let width = CVPixelBufferGetWidth(buffer)
        let height = CVPixelBufferGetHeight(buffer)
        let sourceBytesPerRow = CVPixelBufferGetBytesPerRow(buffer)
        let destinationChannelCount = 3
        let destinationBytesPerRow = destinationChannelCount * width
        
        var sourceBuffer = vImage_Buffer(data: sourceData,
                                         height: vImagePixelCount(height),
                                         width: vImagePixelCount(width),
                                         rowBytes: sourceBytesPerRow)
        
        guard let destinationData = malloc(height * destinationBytesPerRow) else {
            print("Error: tran bo dem")
            return nil
        }
        
        defer {
            free(destinationData)
        }
        
        var destinationBuffer = vImage_Buffer(data: destinationData,
                                              height: vImagePixelCount(height),
                                              width: vImagePixelCount(width),
                                              rowBytes: destinationBytesPerRow)
        
        let pixelBufferFormat = CVPixelBufferGetPixelFormatType(buffer)
        
        switch (pixelBufferFormat) {
        case kCVPixelFormatType_32BGRA:
            vImageConvert_BGRA8888toRGB888(&sourceBuffer, &destinationBuffer, UInt32(kvImageNoFlags))
        case kCVPixelFormatType_32ARGB:
            vImageConvert_ARGB8888toRGB888(&sourceBuffer, &destinationBuffer, UInt32(kvImageNoFlags))
        case kCVPixelFormatType_32RGBA:
            vImageConvert_RGBA8888toRGB888(&sourceBuffer, &destinationBuffer, UInt32(kvImageNoFlags))
        default:
            // Unknown pixel format.
            return nil
        }
        
        let byteData = Data(bytes: destinationBuffer.data, count: destinationBuffer.rowBytes * height)
        if isModelQuantized {
            return byteData
        }
        
        // Not quantized, convert to floats
        let bytes = Array<UInt8>(unsafeData: byteData)!
        var floats = [Float]()
        for i in 0..<bytes.count {
            if self.subtract != 0 {
            floats.append((Float(bytes[i]) - subtract) / subtract)
            } else {
                floats.append((Float(bytes[i])))
            }
           
        }
        return Data(copyingBufferOf: floats)
    }}

// MARK: - Extensions

extension Data {
    init<T>(copyingBufferOf array: [T]) {
        self = array.withUnsafeBufferPointer(Data.init)
    }
}

extension Array {
    init?(unsafeData: Data) {
        guard unsafeData.count % MemoryLayout<Element>.stride == 0 else { return nil }
        #if swift(>=5.0)
        self = unsafeData.withUnsafeBytes { .init($0.bindMemory(to: Element.self)) }
        #else
        self = unsafeData.withUnsafeBytes {
            .init(UnsafeBufferPointer<Element>(
                start: $0,
                count: unsafeData.count / MemoryLayout<Element>.stride
            ))
        }
        #endif  // swift(>=5.0)
    }
}
